@@ -2,45 +2,60 @@
 # vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
+import errno
 import os
 import pwd
 import sys
-import errno
-from collections import namedtuple
 from contextlib import suppress
+from typing import NamedTuple, Optional, Set, TYPE_CHECKING
 
-appname = 'kitty'
-version = (0, 15, 1)
-str_version = '.'.join(map(str, version))
+from .types import run_once
+
+if TYPE_CHECKING:
+    from .options.types import Options
+
+
+class Version(NamedTuple):
+    major: int
+    minor: int
+    patch: int
+
+
+appname: str = 'kitty'
+kitty_face = '🐱'
+version: Version = Version(0, 21, 1)
+str_version: str = '.'.join(map(str, version))
 _plat = sys.platform.lower()
-is_macos = 'darwin' in _plat
-base = os.path.dirname(os.path.abspath(__file__))
+is_macos: bool = 'darwin' in _plat
+if getattr(sys, 'frozen', False):
+    extensions_dir: str = getattr(sys, 'kitty_extensions_dir')
+    kitty_base_dir = os.path.dirname(extensions_dir)
+    if is_macos:
+        kitty_base_dir = os.path.dirname(os.path.dirname(kitty_base_dir))
+    kitty_base_dir = os.path.join(kitty_base_dir, 'kitty')
+else:
+    kitty_base_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
+    extensions_dir = os.path.join(kitty_base_dir, 'kitty')
 
 
-ScreenGeometry = namedtuple('ScreenGeometry', 'xstart ystart xnum ynum dx dy')
-WindowGeometry = namedtuple('WindowGeometry', 'left top right bottom xnum ynum')
+@run_once
+def kitty_exe() -> str:
+    rpath = sys._xoptions.get('bundle_exe_dir')
+    if not rpath:
+        items = os.environ.get('PATH', '').split(os.pathsep) + [os.path.join(kitty_base_dir, 'kitty', 'launcher')]
+        seen: Set[str] = set()
+        for candidate in filter(None, items):
+            if candidate not in seen:
+                seen.add(candidate)
+                if os.access(os.path.join(candidate, 'kitty'), os.X_OK):
+                    rpath = candidate
+                    break
+        else:
+            raise RuntimeError('kitty binary not found')
+    return os.path.join(rpath, 'kitty')
 
 
-def kitty_exe():
-    ans = getattr(kitty_exe, 'ans', None)
-    if ans is None:
-        rpath = sys._xoptions.get('bundle_exe_dir')
-        if not rpath:
-            items = filter(None, os.environ.get('PATH', '').split(os.pathsep))
-            seen = set()
-            for candidate in items:
-                if candidate not in seen:
-                    seen.add(candidate)
-                    if os.access(os.path.join(candidate, 'kitty'), os.X_OK):
-                        rpath = candidate
-                        break
-            else:
-                rpath = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'launcher')
-        ans = kitty_exe.ans = os.path.join(rpath, 'kitty')
-    return ans
-
-
-def _get_config_dir():
+def _get_config_dir() -> str:
     if 'KITTY_CONFIG_DIRECTORY' in os.environ:
         return os.path.abspath(os.path.expanduser(os.environ['KITTY_CONFIG_DIRECTORY']))
 
@@ -58,12 +73,12 @@ def _get_config_dir():
             if os.access(q, os.W_OK) and os.path.exists(os.path.join(q, 'kitty.conf')):
                 return q
 
-    def make_tmp_conf():
+    def make_tmp_conf() -> None:
         import tempfile
         import atexit
         ans = tempfile.mkdtemp(prefix='kitty-conf-')
 
-        def cleanup():
+        def cleanup() -> None:
             import shutil
             with suppress(Exception):
                 shutil.rmtree(ans)
@@ -89,7 +104,8 @@ del _get_config_dir
 defconf = os.path.join(config_dir, 'kitty.conf')
 
 
-def _get_cache_dir():
+@run_once
+def cache_dir() -> str:
     if 'KITTY_CACHE_DIRECTORY' in os.environ:
         candidate = os.path.abspath(os.environ['KITTY_CACHE_DIRECTORY'])
     elif is_macos:
@@ -101,32 +117,16 @@ def _get_cache_dir():
     return candidate
 
 
-def cache_dir():
-    ans = getattr(cache_dir, 'ans', None)
-    if ans is None:
-        ans = cache_dir.ans = _get_cache_dir()
-    return ans
+def wakeup() -> None:
+    from .fast_data_types import get_boss
+    b = get_boss()
+    if b is not None:
+        b.child_monitor.wakeup()
 
 
-def get_boss():
-    return get_boss.boss
-
-
-def set_boss(m):
-    from .fast_data_types import set_boss as set_c_boss
-    get_boss.boss = m
-    set_c_boss(m)
-
-
-def wakeup():
-    get_boss.boss.child_monitor.wakeup()
-
-
-base_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
-terminfo_dir = os.path.join(base_dir, 'terminfo')
-logo_data_file = os.path.join(base_dir, 'logo', 'kitty.rgba')
-logo_png_file = os.path.join(base_dir, 'logo', 'kitty.png')
-beam_cursor_data_file = os.path.join(base_dir, 'logo', 'beam-cursor.png')
+terminfo_dir = os.path.join(kitty_base_dir, 'terminfo')
+logo_png_file = os.path.join(kitty_base_dir, 'logo', 'kitty.png')
+beam_cursor_data_file = os.path.join(kitty_base_dir, 'logo', 'beam-cursor.png')
 try:
     shell_path = pwd.getpwuid(os.geteuid()).pw_shell or '/bin/sh'
 except KeyError:
@@ -135,11 +135,12 @@ except KeyError:
     shell_path = '/bin/sh'
 
 
-def glfw_path(module):
-    return os.path.join(base, 'glfw-{}.so'.format(module))
+def glfw_path(module: str) -> str:
+    prefix = 'kitty.' if getattr(sys, 'frozen', False) else ''
+    return os.path.join(extensions_dir, f'{prefix}glfw-{module}.so')
 
 
-def detect_if_wayland_ok():
+def detect_if_wayland_ok() -> bool:
     if 'WAYLAND_DISPLAY' not in os.environ:
         return False
     if 'KITTY_DISABLE_WAYLAND' in os.environ:
@@ -147,33 +148,14 @@ def detect_if_wayland_ok():
     wayland = glfw_path('wayland')
     if not os.path.exists(wayland):
         return False
-    # GNOME does not support xdg-decorations
-    # https://gitlab.gnome.org/GNOME/mutter/issues/217
-    import ctypes
-    lib = ctypes.CDLL(wayland)
-    check = lib.glfwWaylandCheckForServerSideDecorations
-    check.restype = ctypes.c_char_p
-    check.argtypes = ()
-    try:
-        ans = bytes(check())
-    except Exception:
-        return False
-    if ans == b'NO':
-        print(
-                'Your Wayland compositor does not support server side window decorations,'
-                ' disabling Wayland. You can force Wayland support using the'
-                ' linux_display_server option in kitty.conf'
-                ' See https://drewdevault.com/2018/01/27/Sway-and-client-side-decorations.html'
-                ' for more information.',
-                file=sys.stderr)
-    return ans == b'YES'
+    return True
 
 
-def is_wayland(opts=None):
+def is_wayland(opts: Optional['Options'] = None) -> bool:
     if is_macos:
         return False
     if opts is None:
-        return is_wayland.ans
+        return bool(getattr(is_wayland, 'ans'))
     if opts.linux_display_server == 'auto':
         ans = detect_if_wayland_ok()
     else:
@@ -183,3 +165,24 @@ def is_wayland(opts=None):
 
 
 supports_primary_selection = not is_macos
+
+
+def running_in_kitty(set_val: Optional[bool] = None) -> bool:
+    if set_val is not None:
+        setattr(running_in_kitty, 'ans', set_val)
+    return bool(getattr(running_in_kitty, 'ans', False))
+
+
+def resolve_custom_file(path: str) -> str:
+    path = os.path.expandvars(os.path.expanduser(path))
+    if not os.path.isabs(path):
+        path = os.path.join(config_dir, path)
+    return path
+
+
+def read_kitty_resource(name: str) -> bytes:
+    try:
+        from importlib.resources import read_binary
+    except ImportError:
+        from importlib_resources import read_binary  # type: ignore
+    return read_binary('kitty', name)
